/* oms_qm.h								*/
/* Opnet Model Support package			*/
/* Queue Management sub-package			*/
/* Editted					*/
/****************************************/
/*  Copyright (c) 1987 - 1998			*/
/*            by MIL 3, Inc.           	*/
/*       (A Delaware Corporation)      	*/
/*    3400 International Drive, N.W.   	*/
/*       Washington, D.C., U.S.A.      	*/
/*       All Rights Reserved.          	*/
/****************************************/


#include "oms_bgutil.h"
#include "ip_dgram_sup.h"
#include "ip_addr_v4.h"
#include "rsvp.h"
/* Protect against multiple includes. 	*/
#ifndef	_OMS_QM_H_INCLUDED_
#define _OMS_QM_H_INCLUDED_

#ifndef OMSC_EXPORT
#define OMSC_EXPORT
#endif


/* ---- (SECTION_SC) SYMBOLIC CONSTANTS ---- */

/* Definition of an enumerated type for the Type of Service.	*/
typedef enum OmsT_Qm_Tos 
	{
	OmsC_Qm_Tos_Unspecified					= -1,    
	OmsC_Qm_Best_Effort						= 0,
	OmsC_Qm_Background						= 1,
	OmsC_Qm_Standart						= 2,
	OmsC_Qm_Excellent_Effort				= 3,
	OmsC_Qm_Streaming_Multimedia			= 4,
	OmsC_Qm_Interactive_Multimedia			= 5,
	OmsC_Qm_Interactive_Voice				= 6,
	OmsC_Qm_Tos_Reserved					= 7
	} OmsT_Qm_Tos;


/* OMS QM Package Macro Definitions.	*/
#define		LTRACE_QM						(op_prg_odb_ltrace_active("ip_qos"))
#define		SUCCESS_ENQUEUED				Boolean
#define    	OmsC_Unassigned					-1
#define		OmsC_Uncomputed					-1
#define		OmsC_No_Red						0
#define     OmsC_Red                       	1
#define		OmsC_Wred						2
#define     OmsC_Rsvp_Total_Byte_Count      200000

/* Define a queuing scheme on an interface.	*/
typedef enum
	{
	IpC_No_Queuing,
	IpC_FIFO_Queuing,
	IpC_WFQ_Queuing,
	IpC_Priority_Queuing,
	IpC_Custom_Queuing,
	IpC_VC_Queuing, 
	IpC_DRR_Queuing
	} IpT_Queuing_Scheme;

/* ---- (SECTION_DT) DATA TYPES ---- */

/* This structure defines one queue. It contains some information about its	*/
/* characteristics and above all it stores the packets in a list.			*/
typedef struct
	{
	List *				pk_list_ptr;					/* contains the packets.			*/ 
	double				current_queue_size_in_bytes;	/* Current queue size in bytes.		*/
	double				previous_time; 					/* Previous finish time.			*/
	double				finish_time;					/* finish_time of the packet.		*/
	double				average_queue_size;				/* Calculated for RED.				*/
	double				start_idle_time;				/* Time since when the queue idle.	*/
	Boolean				queue_to_delete;				/* Specify that RSVP will delete the*/
														/* queue when empty.				*/
	} OmsT_Qm_Queue_Info;


/* These structure defines classification parameters for a queue.	*/
typedef struct
	{
   	double		weight;					/* Weight of the queue for WFQ.					*/
	int			byte_count_number;		/* Byte Count Number for Custom Queuing.		*/
	double		max_size;				/* Maximum size of the queue during congestion. */

	/* VC parameters.	*/
	double		arrival_rate;           /* Arrival Rate of flow in each queue for VC */
	double 		virtual_clock;			/* Virtual clock of each flow.				 */
	double		aux_virtual_clock;		/* Auxilary virtual clock of each flow		 */	
	double 		curr_virtual_clock;		/* keeps track of lowest stamp value of packets in each queue */
	Boolean		first_packet_flag;		/* Gets true when first packet of each flow gets to the its   */
	
									    /* queue.													  */
	/* DRR Parameters	*/
	double quantum;
	double d_counter;
	/* RSVP parameters.	*/
	double		rsvp_reserved_bandwidth;	/* Bandwidth allocated by RSVP for this queue.	*/

	/* RED parameters.	*/	
	int				red_status;					/* To enable or disable RED or WRED.		*/
	int				exponential_weight_factor;	/* Parameter used to compute the average	*/
												/* queue size.								*/
	int 			minimum_threshold;			/* When the average queue size is bigger	*/
												/* than the minimum threshold, packets are	*/
												/* randomly dropped.						*/
	int				maximum_threshold;			/* When the average queue size is bigger 	*/
												/* than the maximum threshold, all traffic	*/
												/* is dropped.								*/
	int				mark_probability_denominator;	/* Fraction of the traffic dropped when */
													/* the average queue size reaches the	*/
													/* maximum threshold.					*/
	} OmsT_Qm_Queue_Configuration;


/* Criteria to put an incoming packet into the corresponding queue.		*/
/* Many of these structures can be used as many criteria for one queue.	*/
typedef struct
	{
	int				queue_id;					/* Queue for which the criteria are applied.	*/
	OmsT_Qm_Tos		tos;						/* Type of Service of the packet.		*/
	int				protocol;					/* Protocol of the packet.				*/
	IpT_Address		source_address;				/* Source_address of the packet.		*/
	IpT_Address		dest_address;				/* Destination address of the packet.	*/
	int				source_port;				/* Source port of the packet.			*/
	int				dest_port;					/* Destination port of the packet.		*/
	int				incoming_iface;				/* Incoming interface of the packet.	*/
	} OmsT_Qm_Classification_Criteria;


/* Contains general queuing information about an interface.		*/
/* The queuing attributes of each interface is indentified by	*/
/* a profile name. Each profile	can list many queues and each 	*/
/* queue can have a list of	classification criteria. 			*/	
typedef struct
	{
	char *							name;						/* Name of the queuing scheme.			*/
	int								no_queues;					/* Number of queues.					*/
	int								max_total_no_buf_pkts;		/* Maximum number of buffered packets.	*/
	OmsT_Qm_Queue_Configuration **	queue_configuration;		/* Queue parameters stored as an array	*/
																/* Each index relates to a queue		*/
	List * 							classification_list_ptr;	/* Criteria to classify packets.		*/
	} OmsT_Qm_Attributes;


/* This structure contains all the global informations about the processor, the queue	*/
/* and especially the array of lists of packets.										*/
typedef struct
	{
	OmsT_Qm_Queue_Info **	queue_info_ptr;			/* Pointer to a queue.					*/
	OmsT_Qm_Attributes *	attributes_ptr;			/* Pointer to the attributes.			*/
	int						curr_total_no_buf_pkts;	/* current number of buffered packets.	*/
	double					service_rate;			/* Service rate of the processor.		*/
	int						queue_to_service;		/* Queue which is currently serviced.	*/
	double					total_weight;			/* Total weight of all the queues.		*/
	double					last_real_time;			/* Time of last event (arrival or send)	*/
	double					last_virtual_time;		/* Virtual time of last event.			*/
	int						total_active_weight;	/* Sum of active queues weights.		*/
	double					max_buffer_size;		/* Maximum buffer in bytes on interface.*/
	double					current_buffer_size;	/* Current buffer in bytes on interface.*/
	Boolean					rsvp_flag;				/* Indicate whether RSVP is enabled.	*/
	} OmsT_Qm_Info;


/* This structure defines the main characteristics of a packet so that it should be sent	*/
/* to the correct queue.																	*/
typedef struct
	{
	OmsT_Qm_Tos		tos;					/* Type of Service of the packet.		*/
	int				protocol;				/* protocol of the packet.				*/
	IpT_Address		source_address;			/* source_address of the packet.		*/
	IpT_Address		dest_address;			/* destination address of the packet.	*/
	int				source_port;			/* source port of the packet.			*/
	int				dest_port;				/* destination port of the packet.		*/
	int				incoming_iface;			/* Incoming interface for the packet.	*/
	int				queue_id;				/* queue id where is put the packet.	*/
	} OmsT_Qm_Pkt_Info;	


/* This structure stores all the stathandles for one queue in one interface.	*/
typedef struct
	{
	Boolean			stat_is_registered;					/* Indicates whether the stat is registered.		*/
	Stathandle		buffer_usage_in_pk_stathandle;		/* Statistic for buffer usage in packets.			*/
    Stathandle      buffer_usage_in_bytes_stathandle;   /* Statistic for buffer usage in bytes.             */
	Stathandle		queuing_delay_stathandle;			/* Statistic for queuing delay.						*/
	Stathandle		traffic_dropped_in_pps_stathandle;	/* Statistic for traffic dropped in packets/sec.	*/
	Stathandle		traffic_sent_in_pps_stathandle;		/* Statistic for traffic sent in packets/sec.		*/
	Stathandle		traffic_received_in_pps_stathandle;	/* Statistic for traffic received in packets/sec.	*/
	Stathandle		traffic_dropped_in_bps_stathandle;	/* Statistic for traffic dropped in bits/sec.		*/
	Stathandle		traffic_sent_in_bps_stathandle;		/* Statistic for traffic sent in bits/sec.			*/
	Stathandle		traffic_received_in_bps_stathandle;	/* Statistic for traffic received in bits/sec.		*/
	Stathandle		red_average_queue_size_stathandle;	/* Average queue size calculated by RED.			*/
	Stathandle		virtual_clock_stathandle;		/* Statistic for virtual clock						*/
	Stathandle		a_virtual_clock_stathandle;	/* Statistic for  Auxilary Virtual Clock			*/	 
	Stathandle		curr_virtual_clock_stathandle;		/* Statistic for current Virtual Clock				*/
	Stathandle		virt_clock_stamp_stathandle;
	Stathandle		a_virt_clock_stamp_stathandle;
	Stathandle		deficit_counter_stathandle;
	Stathandle		quantum_stathandle;
	} OmsT_Qm_Stat_Info;


/* These are definitions of function pointer to specify the protocol used : IP, ATM,  ...	*/
typedef OmsT_Qm_Pkt_Info (* OmsT_Qm_Prio_Classifier_Func_Ptr)(Packet * ,OmsT_Qm_Info * ,Boolean *);
typedef OmsT_Qm_Info * (* OmsT_Qm_Info_Assign_Func_Ptr)(int ,double ,OmsT_Qm_Info *);


/* This structure is shared by the root process and the child process	*/
/* in charge of the output IP queuing mechanism.						*/
typedef struct
	{
	List *									iface_table_ptr;				/* List of the connected interfaces.			*/
	int										outstream;						/* Outstream for the packet.					*/
	int										iface_index;					/* Output interface index.						*/
	Stathandle *							locl_pk_dropped_hdl_ptr;		/* Statisitic handle for local packets dropped.	*/
	Stathandle *							globl_pk_dropped_hdl_ptr;		/* Statistic handle for global packets dropped.	*/    
	Stathandle *							locl_num_pkts_sent_hdl_ptr;		/* Statistic handle for total packets sent.		*/
	Stathandle *							locl_num_mcasts_sent_hdl_ptr;	/* Statistic handle for multicast packets sent.	*/
	Stathandle *							locl_num_bcasts_sent_hdl_ptr;	/* Statistic handle for Broadcast packats sent.	*/
	OmsT_Bgutil_Routed_State *				sent_bgutil_state_ptr;			/* Needed to write stats for bgutil.			*/
	double									sent_pk_last_update_time;		/* Needed to write stats for bgutil.			*/
	Boolean									bgutil;							/* To know if bgutil is enabled or not.			*/
	int										statistic_index;				/* index to register queuing statistics.		*/
	int										rsvp_statistic_index;			/* Index to register RSVP statistics.			*/
	RsvpT_TC_Ici_Struct *					rsvp_request_ptr;				/* RSVP request. (OPC_NIL is not used)			*/
	} OmsT_Qm_Shared_Memory;


/* This structure stores CAR parameters for a class of service.	*/
/* A class of service (COS) allows to classify packet and apply	*/
/* a specific action to each of them.							*/
typedef struct
	{
	int						application_type;	/* TCP/UDP protocol type (e.g. 80 for http)		*/
	int						tos;				/* IP precedence in TOS filed.					*/
	int						incoming_port;		/* Interface number on which a packet arrives.	*/
	double					average_rate;		/* Data rate contract.							*/
	double					normal_burst_size;	/* Normal burst size allowed before some 		*/
												/* traffic doesn't comply the policy.			*/
	double					excess_burst_size;	/* Excess burst size allowed before ALL traffic	*/
												/* doesn't comply any more.						*/
	int						conform_precedence;	/* TOS set in conforming IP datagrams.			*/
	int						conform_action;		/* Send the packet or continue the CAR look-up.	*/ 
	int						excess_precedence;	/* TOS set in non-conforming IP datagrams.		*/
	int						excess_action;		/* Drop the packet or continue the CAR look-up.	*/
	} OmsT_Qm_Cos_Definition;


/* Defines a structure for a CAR profile. An array of this structure	*/
/* is created for each interface.										*/
typedef struct
	{
	Boolean							car_status;				/* Status Enabled/Disabled of CAR on the interface.			*/
	char *							name;					/* Name of the CAR profile.									*/	
	int								number_of_cos;			/* Number of class of service  (COS) on this interface.		*/
	OmsT_Qm_Cos_Definition*			cos_definition;			/* Definition of each class of service on this interface.	*/
	} OmsT_Qm_Car_Profile;


/* Structure maintaining the CAR state information for one class of service (COS).	*/
typedef struct
	{
	double			bucket_size;		/* Contains the current bucket size.		*/	
	double			last_update_time;	/* Last time the bucket size was updated.	*/
	double			actual_debt;		/* Current debt of the COS.					*/ 
	double			compounded_debt;	/* Current compounded debt of the COS.		*/
	} OmsT_Qm_Car_Information;


/* Structure storing all the stathandles for CAR statistics on one interface.	*/
typedef struct
	{
	Stathandle		in_traffic_dropped_in_pps_stathandle;
	Stathandle		in_traffic_dropped_in_bps_stathandle;
	Stathandle		out_traffic_dropped_in_pps_stathandle;
	Stathandle		out_traffic_dropped_in_bps_stathandle;
	} OmsT_Qm_Car_Stat_Info;


/* ---- (SECTION_FP) FUNCTION PROTOTYPES ---- */

OMSC_EXPORT char *				oms_qm_tos_value_to_tos_name_convert (OmsT_Qm_Tos tos_value);
OMSC_EXPORT OmsT_Qm_Info *		Oms_Qm_Info_Create (int interface_index, double speed, OmsT_Qm_Info_Assign_Func_Ptr
									assign_function_ptr);
OMSC_EXPORT SUCCESS_ENQUEUED	Oms_Qm_Packet_Enqueue (Packet * pkptr, OmsT_Qm_Info * qm_info, 
									OmsT_Qm_Prio_Classifier_Func_Ptr obtain_priority_classifier, Boolean * 
									warning_message_print_ptr, int * queue_id); 
OMSC_EXPORT int					Oms_Qm_Next_Wfq_Queue_Select (OmsT_Qm_Info * wfq_info);
OMSC_EXPORT int                 Oms_Qm_Next_Pq_Queue_Select (OmsT_Qm_Info * qm_info);
OMSC_EXPORT int                 Oms_Qm_Next_Cq_Queue_Select (OmsT_Qm_Info * qm_info, int queue_id, int * byte_count);
OMSC_EXPORT double				Oms_Qm_Service_Time_Get (int queue_id, OmsT_Qm_Info * qm_info);
OMSC_EXPORT Packet*				Oms_Qm_Packet_Extract (int queue_id, OmsT_Qm_Info * qm_info);
OMSC_EXPORT double              oms_qm_finish_time_compute (int queue_id, OmsT_Qm_Info * qm_info);
OMSC_EXPORT int                 Oms_Qm_Average_Queue_Size_Update (OmsT_Qm_Info * qm_info, int queue_id);
OMSC_EXPORT Boolean             oms_qm_red_packet_drop (OmsT_Qm_Info * qm_info, int queue_id, int tos);
OMSC_EXPORT int					Oms_Qm_Queue_Size_Get (int queue_id, OmsT_Qm_Info * qm_info);
OMSC_EXPORT int					Oms_Qm_Queues_Num_Get (OmsT_Qm_Info * qm_info);
OMSC_EXPORT OmsT_Qm_Info *		ip_qos_info_assign (int interface_index, double speed, OmsT_Qm_Info * qm_info);
OMSC_EXPORT OmsT_Qm_Pkt_Info	ip_qos_tos_classifier_obtain (Packet * pkptr, OmsT_Qm_Info * qm_info,
									Boolean * ip_warning_message_print_ptr);
OMSC_EXPORT OmsT_Qm_Pkt_Info    ip_qos_fifo_classifier_obtain (Packet * pkptr, OmsT_Qm_Info * qm_info,
									Boolean * ip_warning_message_print_ptr);
OMSC_EXPORT Boolean				Ip_Qos_Car_Policy_Limit (Packet* pkptr, int incoming_interface, 
									OmsT_Qm_Car_Profile* car_profile, OmsT_Qm_Car_Information* car_info);
OMSC_EXPORT void				ip_qos_application_type_get (Packet* pkptr, IpT_Dgram_Fields* pk_fd_ptr, int* source_port, int* dest_port);
OMSC_EXPORT OmsT_Qm_Info *      ip_qos_with_rsvp_info_assign (int interface_index, double speed, OmsT_Qm_Info * qm_info);



/* End if for protection against multiple includes. */
#endif /* _OMS_QM_H_INCLUDED_ */

